/***************************************************************
 * Name:      audio.h
 * Purpose:   audio streams raw waveforms to audio
 * Author:    John Conner Atkeson
 * Created:   2007-11-06
 * Copyright: John Atkeson (http://jca3.freeshell.org)
 * License:   LPGL
 *  code heavily derived from:
 *  https://lampka.siedziba.pl/WebSVN/filedetails.php?repname=siedziba+svn&path=%2FSound_Test%2FSoundTest.cpp&rev=256&sc=1
 **************************************************************/
#include "audio.h"
// https://lampka.siedziba.pl/WebSVN/filedetails.php?repname=siedziba+svn&path=%2FSound_Test%2FSoundTest.cpp&rev=256&sc=1
#include "ctype.h"
#include "SDL/SDL.h"//-lSDLmain -lSDL -mwindow
#include "SDL/SDL_audio.h"
#include <math.h>
#include <stdio.h>
#ifdef WINDOWS
#include "windows.h"
#endif

#ifdef WINDOWS
#define die(...) { char err[100]; snprintf(err, sizeof(err), __VA_ARGS__); MessageBox(NULL, err, "Error", MB_OK | MB_ICONEXCLAMATION); exit(1); }
#else
#define die(...) { fprintf(stderr, __VA_ARGS__); fprintf(stderr, "\n"); exit(1); }
#endif
#include <windows.h>

#include <pthread.h>

/*  Make the class name into a global variable  */
// char szClassName[ ] = "CodeBlocksWindowsApp";

//unsigned             Initialised = 0;

// User controllable variables
// const unsigned  SAMPLE_FREQ = 44100;     // sampling frequency
// unsigned        sdl_samples = 2048;          // SDL buffer size

/* ************************************************************************************************************ */
/* Playable_Class is the base class for every object that can output a waveform. */
Playable_Class::Playable_Class(){
  time=0.0; this->set_octave(7.0); loudness=1.0;
};
void Playable_Class::set_octave(double new_octave){
  octave = new_octave;
  frequency = Math.Pow(2.0, new_octave);
};
void Playable_Class::set_frequency(double new_frequency){
  octave = Math.Log(new_frequency)/Math.Log(2.0);
  frequency = new_frequency;
};
double Playable_Class::getamp(double time){
  return 0.0;
};
double Playable_Class::maxamp(){
  return 1.0;
};
double Playable_Class::get_duration(){// tells how long this track will play
  return 1.0;
};
bool Playable_Class::keepgoing(){// tells when we have passed the end of duration.
  return true;
};
class Time_Slicer_Class* Playable_Class::spawn_slicer(){// creates a "read head" to keep a time marker in this playable
  return new Time_Slicer_Class(this);
};

/* **********************************************************************************************************************
  Begin Time_Slicer_Class functions.
********************************************************************************************************************** */
Time_Slicer_Class::Time_Slicer_Class(class Playable_Class *new_playable){
  my_playable = new_playable;
  time_abs=0.0; octave_abs=0.0; loudness_abs=0.0;
  parent = NULL;
  keeprunning = true;
};
//-------------------------------------------------------------------------------------
Time_Slicer_Class::~Time_Slicer_Class(){};
//-------------------------------------------------------------------------------------
void Time_Slicer_Class::init_slicer(double start_time, int new_start_dex, int new_end_dex){// this one is different for every descendent
};
//-------------------------------------------------------------------------------------
double Time_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  double returnval = 0.0;
  nowtime -= my_playable->time; nowoctave += my_playable->octave;/* yloc is base octave for this frame */ nowvol *= my_playable->loudness;
  returnval = my_playable->getamp(nowtime);
  return returnval;
};
//-------------------------------------------------------------------------------------
bool Time_Slicer_Class::keepgoing(){// tells when we have passed the end of duration.
  return keeprunning;
};
//-------------------------------------------------------------------------------------
void Time_Slicer_Class::Establish_Location(){// reestablish my absolute coordinates based on parents
  /* This is called: 1. when it is spawned for playing, and 2. whenever my playable or any of its parents are moved. */
  Time_Slicer_Class *pptr = this->parent;
  double xloc, yloc, zloc;
  xloc=my_playable->time; yloc=my_playable->octave; zloc=my_playable->loudness;
  Playable_Class *playable;
  while (pptr!=NULL){// reach up through all tree parents
    playable = pptr->my_playable;
    xloc += playable->time;
    yloc += playable->octave;
    zloc *= playable->loudness;// multiply loudness, envelopes within envelopes
    pptr = pptr->parent;
  }
  this->time_abs=xloc; this->octave_abs=yloc; this->loudness_abs=zloc;
};

/* **********************************************************************************************************************
  Begin Time_Slicer_Tree_Class functions.
********************************************************************************************************************** */
Time_Slicer_Tree_Class::Time_Slicer_Tree_Class(class Playable_Tree_Class *new_playable):Time_Slicer_Class(new_playable){
  subslicers = new ArrayList<Time_Slicer_Class>();
  my_playable_tree = new_playable;
};
//-------------------------------------------------------------------------------------
//overrider void Time_Slicer_Tree_Class::Establish_Location();// reestablish my absolute coordinates based on parents
//-------------------------------------------------------------------------------------
overrider Time_Slicer_Tree_Class::~Time_Slicer_Tree_Class(){
  delete subslicers;
};
//-------------------------------------------------------------------------------------
overrider void Time_Slicer_Tree_Class::init_slicer(double start_time, int new_start_dex, int new_end_dex){// this one is different for every descendent
};
//-------------------------------------------------------------------------------------
overrider double Time_Slicer_Tree_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  double returnval = 0.0;
  nowtime -= my_playable->time; nowoctave += my_playable->octave;/* yloc is base octave for this frame */ nowvol *= my_playable->loudness;
  int num_subslices = subslicers->size();
  for (int scnt=0;scnt<num_subslices;scnt++){
    Time_Slicer_Class *slice = subslicers->get(scnt);
    returnval += slice->getslice(nowtime, nowoctave, nowvol);
  }
  // if we have passed the end of my playable's duration, set keepgoing to false.
  // keepgoing = my_playable->keepgoing();
  return returnval;
};
//-------------------------------------------------------------------------------------
overrider bool Time_Slicer_Tree_Class::keepgoing(){// tells when we have passed the end of duration.
  return keeprunning;
};

/* **********************************************************************************************************************
  Begin Player_Class functions and callback
********************************************************************************************************************** */
void player_callback_wrapper(void *environment, Uint8 *stream, int len){
  Player_Class *player = (Player_Class*)environment;  stereo *fill = (stereo*)stream;
  // player->fill_outbuffer(fill, len/sizeof(stereo));
  player->fill_outbuffer_sgen(fill, len/sizeof(stereo));
};
//-------------------------------------------------------------------------------------
void Player_Class::sdlsoundinit(){
  // initialise the SDL system
  // sdlinit(); return;
  if (false){
    if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_VIDEO) == -1){
      die("Could not initialize SDL: %s", SDL_GetError());
    }
  }else{
    /* Initialization */
    if (SDL_InitSubSystem(SDL_INIT_AUDIO) == -1){
      // SDL Audio subsystem could not be started
      die("Could not initialize SDL: %s", SDL_GetError());
    }
  }
  atexit(SDL_Quit);
  if (Initialised) SDL_CloseAudio();
  SDL_AudioSpec format;
  format.format = AUDIO_S16SYS;
  format.channels = 2;
  format.freq = SAMPLE_FREQ;
  format.samples = sdl_samples;
  format.callback = player_callback_wrapper;
  format.userdata = this;
  if (SDL_OpenAudio(&format, NULL) < 0) die("SDL_OpenAudio failed: %s", SDL_GetError());
  SDL_PauseAudio(0);
  Initialised = 1;
};
//-------------------------------------------------------------------------------------
int Player_Class::fill_outbuffer(stereo *fill, int playsize){// virtual
#if 0
  float sine_step = Frequency * twopi / SAMPLE_FREQ;
  while (playsize > 0) // write into the sound buffer
  {
    CyclePos += sine_step;
    if (CyclePos > twopi) CyclePos -= twopi;
    fill->left = fill->right = (short)nearbyintf(12000 * sinf(CyclePos));
    fill++;
    playsize --;
  }
#endif
  return 0;
};
//-------------------------------------------------------------------------------------
int Player_Class::fill_outbuffer_sgen(stereo *fill, int playsize){
  while (playsize > 0) // write into the sound buffer
  {
    // fill->left = fill->right = (short)nearbyintf(12000 * sgen->getamp(time));
    fill->left = fill->right = (short)nearbyintf(12000 * slicer->getslice(time, 0.0, 1.0));
    time+=timeinc;
    fill++;
    playsize --;
  }
  return 0;
};
#if 0
/*
next step is to figure out how to play a Voice_Track_Class (voice) in a loop.;
that is a subset of the Voice_Track_Class, startdex to enddex.;

at some point we want more than sine waves;
create a gui freewave_generator_class editor.;

ideally the player defines the time range to play in a loop,;
and ideally the current time must only be expressed to the playables as an absolute time.;

how does the player know the end-of-loop time when that time is moving?;
*/
#endif
/* **********************************************************************************************************************
  End Player_Class methods and callback
********************************************************************************************************************** */
/* **********************************************************************************************************************
  Begin Signal_Generator_Class methods
********************************************************************************************************************** */
Signal_Generator_Class::Signal_Generator_Class(){
  loudness = 1.0;
  // freq = 261.626;// default to approximately middle C
  freq = 440.0;// default to A, = octave 8.7813597135246596040696824762296
  clear();
}
//-------------------------------------------------------------------------------------
double Signal_Generator_Class::getamp(double time){// time is in absolute seconds from T=0.
  calc_angle(time);
  return Math.Sin(current_angle * twopi) * loudness;// twopi == 360 degrees (radians)
}
//-------------------------------------------------------------------------------------
double Signal_Generator_Class::maxamp(){
  return 1.0;// should be 1.0
}
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::clear(){
  reset_time(0.0); current_angle = 0.0;
}
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::reset_time(double new_time){
  prev_time = new_time;//
}
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::set_octave(double new_octave){
  octave = new_octave;
  freq = Math.Pow(2.0, new_octave);
};
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::set_frequency(double new_freq){
  freq = new_freq;
  octave = Math.Log(new_freq)/Math.Log(2.0);
  //set_octave(Math.Log(new_freq)/Math.Log(2.0)); // alternative
};
//-------------------------------------------------------------------------------------
void Signal_Generator_Class::calc_angle(double time){
  double delta_time = time-prev_time;
  double delta_angle = delta_time * freq;// 1.0 == 360 degrees (fractions along the way)
  current_angle += delta_angle;// current_angle = current_angle-Math.floor(current_angle);
  prev_time = time;
};
/* **********************************************************************************************************************
  End Signal_Generator_Class methods
********************************************************************************************************************** */

/* **********************************************************************************************************************
  Begin Freewave_Generator_Class methods
********************************************************************************************************************** */
Freewave_Generator_Class::Freewave_Generator_Class(double new_freq){
  rez = 1000;
  //rez = 10000;
  frez = (double)rez;// resolution of our waveform
  ampray = allocsafe(rez, double);
  freq = new_freq;
#if 0
  for (int cnt=0;cnt<rez;cnt++){
    ampray[cnt] = (  2.0*(((double)cnt)/frez)  )-1.0;// sawtooth wave
  }
#else
  for (int cnt=0;cnt<rez;cnt++){
    ampray[cnt] = Math.Sin(twopi*((double)cnt)/frez);// sine wave
  }
#endif
};
//-------------------------------------------------------------------------------------
Freewave_Generator_Class::~Freewave_Generator_Class(){
  freesafe(ampray);
};
//-------------------------------------------------------------------------------------
overrider double Freewave_Generator_Class::getamp(double time){// time is in absolute seconds from T=0.
  calc_angle(time);
  double base_angle = current_angle;
  double base = Math.Floor(base_angle); base_angle -= base;
  base_angle *= frez;

  // get the two array indexes that surround us
  int dex0 = (int)base_angle; // int dex2 = (dex+1) % 100;
  int dex1 = (dex0+1); if (dex1>=rez) {
    dex1=0;
  }// wrap to beginning

  // interpolate between points.
  double weight = base_angle; base = Math.Floor(weight); weight -= base;
  double value = (ampray[dex0]*(1.0-weight)) + (ampray[dex1]*weight);
  return value * loudness;
};
#if 0
//-------------------------------------------------------------------------------------
double Freewave_Generator_Class::find_by_time(double time){// time is in absolute seconds from T=0.
  calc_angle(time);
  double base_angle = current_angle;
  double base = Math.Floor(base_angle); base_angle -= base;
  base_angle *= frez;

  // get the two array indexes that surround us
  int dex0 = (int)base_angle; // int dex2 = (dex+1) % 100;
  int dex1 = (dex0+1);
  if (dex1>=rez) {
    dex1=0;
  }// wrap to beginning

  // interpolate between points.
  double weight = base_angle; base = Math.Floor(weight); weight -= base;
  double value = (ampray[dex0]*weight) + (ampray[dex1]*(1.0-weight));
  return value * loudness;
};
#endif

/* **********************************************************************************************************************
  Begin Splinewave_Generator_Class methods
********************************************************************************************************************** */
Splinewave_Generator_Class::Splinewave_Generator_Class(double new_freq):Freewave_Generator_Class(new_freq)
{
  Point_Class *pnt;
  spline = new ArrayList<Point_Class>();
  // xp0 = 2.2; xp1 = 12.3; xp2 = 15.6; yp0 = 2.2; yp1 = 12.3; yp2 = 15.6;
  switch (4){
  case 0:{
    pnt = new Point_Class(0.22,  0.22); spline->add(pnt);
    pnt = new Point_Class(0.46, -0.90); spline->add(pnt);
    pnt = new Point_Class(0.53,  0.93); spline->add(pnt);
    pnt = new Point_Class(0.76, -0.40); spline->add(pnt);
    break;
  }
  case 1:{// pointy wave
    pnt = new Point_Class(0.22, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.45,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.55,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 2:{// has gaps
    pnt = new Point_Class(0.20, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.40,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.60,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 3:{// symmetric diving bell
    pnt = new Point_Class(0.20, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.30,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.70,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 4:{
    pnt = new Point_Class(0.01, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.30,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.50,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.70,  0.99); spline->add(pnt);
    pnt = new Point_Class(0.75, -0.99); spline->add(pnt);
    pnt = new Point_Class(0.80, -0.99); spline->add(pnt);
    break;
  }
  case 5:{//senseless loop
    double num=20;
    for (double cnt=0;cnt<num;cnt++){
      pnt = new Point_Class(cnt/num, 0.0); spline->add(pnt);
    }
    break;
  }
  }/* switch */
  apply_spline();
};
//-------------------------------------------------------------------------------------
void Splinewave_Generator_Class::apply_spline(){// apply the spline to the amplitude array
  Point_Class *first_point, *prev_point, *next_point;
  int num_points = spline->size();

  first_point = spline->get(num_points-2);// last_point-1

  double xfirst,yfirst;
  xfirst = first_point->xloc-1.0; yfirst = first_point->yloc;// wrap

  double xprev,yprev, xnext,ynext;
  prev_point = spline->get(num_points-1);// last_point
  xprev = prev_point->xloc-1.0; yprev = prev_point->yloc;// wrap

  double xcut_prev, ycut_prev, xcut_next, ycut_next;
  xcut_prev = (xfirst+xprev)/2; ycut_prev = (yfirst+yprev)/2;
  xcut_next = (xprev+xnext)/2; ycut_next = (yprev+ynext)/2;

  double xtime=0.0;
  double yval=0.0;

  int timedex = (int)(xcut_prev*frez);// increment by integer to avoid cumulative error
  int raydex = (int)((xcut_prev-Math.Floor(xcut_prev))*frez);
  xtime = ((double)timedex)/frez;
  for (int pcnt=0;pcnt<num_points;pcnt++){
    next_point = spline->get(pcnt); xnext = next_point->xloc; ynext = next_point->yloc;
    xcut_next = (xprev+xnext)/2; ycut_next = (yprev+ynext)/2;
    while (xtime<=xcut_next){
      yval = spline_amp(xtime, xcut_prev,ycut_prev, xprev,yprev, xcut_next,ycut_next);
      this->ampray[raydex % rez] = yval;
      raydex++;
      timedex++; xtime = ((double)timedex)/frez;
    }
    xcut_prev = xcut_next; ycut_prev = ycut_next;
    xprev = xnext; yprev = ynext;
  }
};/* apply_spline */
//-------------------------------------------------------------------------------------
double Splinewave_Generator_Class::spline_amp(double X, double xp0, double yp0, double xp1, double yp1, double xp2, double yp2)
{// spline algerbra computed by http://www.quickmath.com/
  double Y, t, comp_t, numer, denom;
  denom = (xp0 + xp2 - (2.0 * xp1));
  if (denom==0.0) {
    xp0+=fudge;// xp0+=0.00000000000001;
    denom = fudge;
  }
  numer = (xp0 - xp1 + Math.Sqrt((xp1 * xp1) - (2.0 * X * xp1) - (xp0 * xp2) + (xp0 * X) + (xp2 * X)));
  t = numer / denom;
  comp_t = 1.0 - t;
  Y = (comp_t * comp_t * yp0) + (2.0 * t * comp_t * yp1) + (t * t * yp2);
  return Y;
};
/* **********************************************************************************************************************
  Begin Voice_Track_Class methods
********************************************************************************************************************** */
Voice_Track_Class::Voice_Track_Class(Signal_Generator_Class *new_sgen, Player_Class *new_audio){// constructor
  my_audio = new_audio;
  length = 0.0;
  // sgen = new Signal_Generator_Class();
  sgen = new_sgen;//new Freewave_Generator_Class(220.0);
  // I guess we invoke this in the drawing_canvas;
  //fill_test();
};
//-------------------------------------------------------------------------------------
Voice_Track_Class::~Voice_Track_Class(){// destructor
  int num_events = this->size();
  for (int cnt=0; cnt<num_events;cnt++){
    Event_Marker_Class *event = this->get(cnt);
    delete event;
  }
  // is voice track responsible for deleting its own signal generator?
  //delete sgen;
};
//-------------------------------------------------------------------------------------
overrider void Voice_Track_Class::add(Event_Marker_Class *new_member){
  ArrayList<Event_Marker_Class>::add(new_member);
  new_member->my_voice = this;// connect back to parent
  //wxMouseEventHandler(Waveform_Canvas_Class::OnDrawingPanelLeftDown));
  //(Voice_Track_Class::)(Event_Marker_Class*) fred = Voice_Track_Class::add;
  // void (Voice_Track_Class::)(Event_Marker_Class*) fred = Voice_Track_Class::add;
  //void *fred = Voice_Track_Class::add;
};
//-------------------------------------------------------------------------------------
int Voice_Track_Class::find_by_time(double target_time){// given the time, treesearch and return the closest index
  int retval = 0;
  int num_events = this->size();
  int last_event = num_events-1;
  Event_Marker_Class *event_prev = this->get(0);
  Event_Marker_Class *event_next = NULL;
  int cnt = 0;
  while (cnt<last_event){
    cnt++;
    event_next = this->get(cnt);
    if (event_next->time > target_time){
      break;
    }
    event_prev = event_next;
  }
  if (event_next!=NULL){
    double diff0 = fabs(event_prev->time - target_time);
    double diff1 = fabs(event_next->time - target_time);
    if (diff0<diff1){// closer to prev
      retval = cnt-1;
    }else{
      retval = cnt;
    }
  }else{
    retval = cnt-1;
  }
  printf("timint:%i\n",retval);
  return retval; // taskkill /F /IM fretless.exe
#if 0 // treesearch
  double targetval, minval, medval, maxval;
  int mindex=0, meddex, maxdex=size();
  minval = this->get(mindex);
  while (mindex<maxdex){
    meddex = (mindex+maxdex)/2; medval = this->get(meddex);
    if (targetval<=medval){
      maxdex = meddex;
    }else{
      mindex = meddex+1; minval = this->get(mindex);
    }
  }
  return mindex;
#endif
};/* find_by_time */
//-------------------------------------------------------------------------------------
void Voice_Track_Class::fill_test(){// fill Voice_Track with dummy values
  Event_Marker_Class *event = new Event_Marker_Class();
  event->time = 0.0;
  //event.set_frequency(440);
  event->set_octave(Math.Log(440)/Math.Log(2.0));
  this->add(event);

  event = new Event_Marker_Class();
  event->time = 0.5;
  //event.set_frequency(220);
  event->set_octave(Math.Log(220)/Math.Log(2.0));
  event->loudness = 0.5;
  this->add(event);

  event = new Event_Marker_Class();
  event->time = 1.0;
  //event.set_frequency(440);
  event->set_octave(Math.Log(440)/Math.Log(2.0));
  // event.loudness = 0.0;
  this->add(event);

  event = new Event_Marker_Class();
  event->time = 1.5;
  // event->frequency = 440;
  event->set_octave(Math.Log(440)/Math.Log(2.0));
  this->add(event);

  event = new Event_Marker_Class();
  event->time = 3.0;
  // event->frequency = 880;
  event->set_octave(Math.Log(880)/Math.Log(2.0));
  event->loudness = 0.0;
  this->add(event);
  length = event->time;
};
//-------------------------------------------------------------------------------------
void Voice_Track_Class::play_segment(int start_dex, int end_dex){
  // morphable_generator_class sgen = new morphable_generator_class();
  sgen->clear();
  Event_Marker_Class *event_prev = this->get(start_dex);
  for (int cnt=start_dex+1;cnt<=end_dex;cnt++){// inclusive
    Event_Marker_Class *event_next = this->get(cnt);
    // System.out.println("Event number: " + cnt);
    for (double time=event_prev->time; time<event_next->time; time+=my_audio->timeinc){
      // event_prev.intep_to_next_frequency(event_next, time, sgen);
      event_prev->intep_to_next_octave(event_next, time, sgen);
      double amp = 0.0;
      amp += sgen->getamp(time);
      // amp *= (2.0/3.0);// tweak to fit into range of this audio card.
      my_audio->add_moment(amp, amp);// left and right stereo, respectively
    }
    event_prev = event_next;
  }
};
//-------------------------------------------------------------------------------------
overrider double Voice_Track_Class::get_duration(){// tells how long this track will play
  return 1.0;
};
overrider bool Voice_Track_Class::keepgoing(){// tells when we have passed the end of duration.
  return false;
  // return keeprunning;
};
//-------------------------------------------------------------------------------------
overrider Time_Slicer_Class* Voice_Track_Class::spawn_slicer(){// creates a "read head" to keep a time marker in this playable
  Voice_Track_Slicer_Class *vs = new Voice_Track_Slicer_Class(this);
  vs->init_slicer(0.0, 0, this->size()-1);
  return vs;
};
//-------------------------------------------------------------------------------------
size_t Voice_Track_Class::Tree_Search(double keytime){// Tree search for the first place where this keytime would go in order.
  size_t minloc=0, maxloc=this->size(), medloc;
  while (maxloc>minloc){
    /* medloc is the average between min and max. */
    medloc=(minloc+maxloc)>>1; /* >>1 is same as div 2, only faster. */
    Event_Marker_Class *cpoint = this->get(medloc);
    if (keytime<=cpoint->time){maxloc=medloc;}/* key is less than or equal to this list entry. */
    else{ minloc=medloc+1; }
  }
  return(minloc);
};/* Tree_Search */

/* cloning?
 http://www.new-brunswick.net/workshop/c++/faq/virtual-functions.html
 Voice_Track_Class* Voice_Track_Class::clone()  const { return new Voice_Track_Class(*this); }
*/
/* **********************************************************************************************************************
  End Voice_Track_Class methods
********************************************************************************************************************** */
//-------------------------------------------------------------------------------------
Voice_Track_Class::Voice_Track_Slicer_Class::Voice_Track_Slicer_Class(Voice_Track_Class *new_timeline):Time_Slicer_Class(new_timeline){
  Voice_Track = new_timeline;
  sgen = Voice_Track->sgen;
  event0_dex=0; event1_dex=0;
};
//-------------------------------------------------------------------------------------
overrider void Voice_Track_Class::Voice_Track_Slicer_Class::init_slicer(double start_time, int new_start_dex, int new_end_dex){
  keeprunning = true;
  int first_cpoint = Voice_Track->Tree_Search(start_time);
  if (first_cpoint<Voice_Track->size()){
  }
  start_dex = new_start_dex; end_dex = new_end_dex;
  event0_dex=start_dex; event1_dex=start_dex+1;
  event0 = Voice_Track->get(event0_dex);
  event1 = Voice_Track->get(event1_dex);
  sgen->clear();
  local_time = 0;// local time inside this block
};
//-------------------------------------------------------------------------------------
overrider double Voice_Track_Class::Voice_Track_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  double returnval = 0.0;
  // lots of effort for a simple loop - simplify please
  double playtime = nowtime - this->my_playable->time;
  if (playtime >= event1->time){// iterators
    if (event1_dex<end_dex){// rollover
      event0 = event1; event0_dex=event1_dex;
      event1_dex++;
      event1 = Voice_Track->get(event1_dex);
    }else{
      // then one full iteration has passed, jump out.
      returnval = 0.0;
      return returnval;
    }
  }
  {// meat of computation
    event0->intep_to_next_octave(event1, playtime, sgen);
    // double amp = event0->intep_to_next_octave(event1, time, sgen);
    returnval = sgen->getamp(playtime);
  }
  return returnval;
};
//-------------------------------------------------------------------------------------
/* Playable_Class members */
/* **********************************************************************************************************************
  Begin Event_Marker_Class methods and callback
********************************************************************************************************************** */
Event_Marker_Class::Event_Marker_Class(){//:Playable_Class(){
  my_voice = NULL;// new_voice;
  this->time=0.0; this->set_frequency(440.0); this->loudness=1.0;;// defaults
  audible = true;// keep playing from this event to the next
};
//-------------------------------------------------------------------------------------
void Event_Marker_Class::intep_to_next_frequency(Event_Marker_Class *nextmarker, double absolutetime, Signal_Generator_Class *sgen){
  double fullspan = nextmarker->time - this->time;
  //if (fullspan==0.0){fullspan=fudge;}// prevent divide-by-zero.
  double delta = absolutetime - this->time;
  double weight_next = delta/fullspan;
  double weight_prev = 1.0 - weight_next;

  double interloudness = (this->loudness * weight_prev) + (nextmarker->loudness * weight_next);
  double interfreq = (this->frequency * weight_prev) + (nextmarker->frequency * weight_next);
  // System.out.println(weight_prev);
  sgen->freq = interfreq; sgen->loudness = interloudness;
  // System.out.println("f:"+interfreq+" t:"+absolutetime);
  // System.out.println(" "+interfreq+", "+absolutetime);
  if (!audible){
    sgen->loudness = 0.0;
  }
};
//-------------------------------------------------------------------------------------
void Event_Marker_Class::intep_to_next_octave(Event_Marker_Class *nextmarker, double absolutetime, Signal_Generator_Class *sgen){
  double fullspan = nextmarker->time - this->time;
  double delta = absolutetime - this->time;
  double weight_next = delta/fullspan;
  double weight_prev = 1.0 - weight_next;
  double interloudness = (this->loudness * weight_prev) + (nextmarker->loudness * weight_next);
  double interoctave = (this->octave * weight_prev) + (nextmarker->octave * weight_next);
  sgen->set_octave(interoctave); sgen->loudness = interloudness;
  if (!audible){
    sgen->loudness = 0.0;
  }
};
overrider double Event_Marker_Class::getamp(double time){// only used for jugar
  Signal_Generator_Class *sgen = this->my_voice->sgen;
  sgen->set_octave(this->octave);
  double amp = sgen->getamp(time) * this->loudness;
  return amp;
};
/* these are all of the playable_class functions, which are only if ever used during jugar */
overrider double Event_Marker_Class::maxamp(){return 0.0;};
overrider double Event_Marker_Class::get_duration(){return 0.0;};// tells how long this track will play
overrider bool Event_Marker_Class::keepgoing(){return 0.0;};// tells when we have passed the end of duration.
overrider class Time_Slicer_Class* Event_Marker_Class::spawn_slicer(){
  return new Event_Marker_Class::Event_Marker_Slicer_Class(this);
};// creates a "read head" to keep a time marker in this playable
/* **********************************************************************************************************************
  End Event_Marker_Class methods and callback
********************************************************************************************************************** */
/* **********************************************************************************************************************
  Begin Event_Marker_Slicer_Class methods and callback
********************************************************************************************************************** */
Event_Marker_Class::Event_Marker_Slicer_Class::Event_Marker_Slicer_Class(class Event_Marker_Class *New_Event_Marker):Time_Slicer_Class(New_Event_Marker){
  My_Event_Marker = New_Event_Marker;
};
overrider void Event_Marker_Class::Event_Marker_Slicer_Class::init_slicer(int new_start_dex, int new_end_dex){// this one is different for every descendent
};
overrider double Event_Marker_Class::Event_Marker_Slicer_Class::getslice(double nowtime, double nowoctave, double nowvol){// Called once per timeslice. Must always go forward.
  // only used for jugar
  //cpoint->spawn_slicer
  //nowtime -= xloc; nowoctave += yloc;/* yloc is base octave for this frame */ nowvol *= my_playable->zloc;
  //nowtime -= my_playable->xloc; nowoctave += my_playable->yloc;/* yloc is base octave for this frame */ nowvol *= my_playable->zloc;
  nowoctave += this->My_Event_Marker->octave;// my_playable->octave;
  Signal_Generator_Class *sgen = this->My_Event_Marker->my_voice->sgen;
  //nowoctave = Math.Log(440)/Math.Log(2.0);
  sgen->set_octave(nowoctave);
  sgen->loudness = this->My_Event_Marker->loudness;
  double amp = sgen->getamp(nowtime) * nowvol;
  return amp;
};
overrider bool Event_Marker_Class::Event_Marker_Slicer_Class::keepgoing(){// tells when we have passed the end of duration.
  return keeprunning;
};
/* **********************************************************************************************************************
  End Event_Marker_Slicer_Class methods and callback
********************************************************************************************************************** */
